"""
Stream tracer running in its own thread
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
import threading


class ThreadedStreamTracer(threading.Thread):
    """Allows tracing a stream in a separate thread

    Properties:
    trace (string) - the aggregated trace
    """
    def __init__(self, source, printer, done_test, line_handler=None):
        """Initializes this instance for tracing

        :param source: (stream) the traced stream
        :param printer: (lambda x) the tracing function
        :param done_test: (lambda) the 'end-of-stream' test
        :param line_handler: (LineHandler) special lines handler
        """
        self.__source = source
        self.__printer = printer
        self.__done_test = done_test
        self.__line_handler = line_handler
        threading.Thread.__init__(self)
        self.trace = []

    def run(self):
        """Performs the trace to the end of the stream
        """
        after_blank = False
        while True:
            line = self._decode_next_line()
            if line == '':
                if self.__done_test():
                    break
                if after_blank:
                    continue
                after_blank = True
            else:
                after_blank = False
            self._trace(line, after_blank)
        self.__source.close()

    def _decode_next_line(self):
        try:
            return self.__source.readline().decode('ascii')
        except UnicodeDecodeError:
            return '<<DECODE FAILED>>'

    def _trace(self, line, after_blank):
        stripped = line.strip()
        if after_blank or not self.__line_handler or not self.__line_handler.handle(line):
            self.__printer(stripped)
        self.trace.append(stripped)



"""
Build events
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""


class BuildEvents:
    """Build events
    """

    def run_pre_build(self):
        """Runs pre-build event
        """
        self._pre_build()

    def run_post_build(self):
        """Runs post-build event
        """
        self._post_build()

    def __init__(self, pre_build=None, post_build=None):
        """Initialize this instance

        :param post_build: post-build action (optional)
        """
        self._pre_build = pre_build if pre_build else self.nothing
        self._post_build = post_build if post_build else self.nothing

    @staticmethod
    def nothing():
        """Empty action
        """



"""
Traced CLI call
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from .call_base import CallBase
from .threaded_tracer import ThreadedStreamTracer


class TracedCall(CallBase):
    """Used to execute shell commands and hold the execution results while tracing to log
    """

    def _run_and_get_streams(self):
        """Traces prints from the process' stdout and stderr using threaded tracers, returning the
         aggregated results

        :return (string, string): (stdout, stderr) strings
        """
        tracers = [ThreadedStreamTracer(self._process.stdout, self._logger.info, self.__done_test),
                   ThreadedStreamTracer(self._process.stderr, self._logger.error, self.__done_test)]
        for tracer in tracers:
            tracer.start()
        for tracer in tracers:
            tracer.join()
        return tuple(['\n'.join(tracer.trace) for tracer in tracers])

    def __done_test(self):
        """Tests if the process has ended

        :return: True if process ended, false otherwise
        """
        return self._process.poll() is not None



"""
Handler for lines
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""


class LineHandler:
    """Handler for lines"""

    def __init__(self, matcher, handler):
        """Initialize this instance

        :param matcher: Predicate for matching lines to this handler
        :param handler: Function for handling matched lines
        """
        self._matcher = matcher
        self._handler = handler

    def handle(self, line):
        """Handle a line

        :param line: Line to handle
        :return: True if line was handled, false otherwise
        """
        if self._matcher(line):
            self._handler(line)
            return True
        return False



"""
Pykrete calls
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""

from .checked_call import CheckedCall
from .traced_call import TracedCall
from .line_handler import LineHandler

__all__ = ['CheckedCall', 'TracedCall', 'LineHandler']



"""
Simple CLI call
Author: Shai Bennathan - shai.bennathan@gmail.com
(C) 2020
"""
from .call_base import CallBase


class CheckedCall(CallBase):
    """Used to execute shell commands and hold the execution results"""

    def __init__(self, cmd):
        """Initializes this instance with the specified command

        :param cmd: (string/list) shell command
        """
        super().__init__(cmd)
        self._logger.debug(
            'Command result:\n--- >> start output << ---\n%s\n--- >> end output << ---',
            self.get_output())

    def _run_and_get_streams(self):
        """
        :return (string, string): (stdout, stderr) strings
        """
        return tuple([stream.decode('ascii').strip() for stream in self._process.communicate()])
